"""Utilities for Pandas."""

from typing import Callable, Dict, Iterator, IO, Optional, Sequence, Tuple, TextIO, Union

from dataclasses import dataclass, field
import os
import sys

try:
    import pandas as pd
except ImportError:
    raise ImportError(
        """
        Pandas not installed. Try installing with pip:
            $ pip install pandas
        or reinstall carabiner with pandas:\n"
            $ pip install carabiner[pd]
        """
    )

from ..io import get_lines
from ..utils import print_err

@dataclass
class IOFormat:

    """Stores delimiters for reading and writing.

    Parameters
    ----------
    in_delim : str
        Delimiter when reading.
    out_delim : str, optional
        Delimiter when writing.
    
    """

    in_delim: str
    out_delim: Optional[str] = None
    strict: bool = True

    def __post_init__(self):
        self.out_delim = self.out_delim or self.in_delim
        if self.in_delim == '\t' and not self.strict:
            self.in_delim = r'\s+'


_FORMAT: Dict[str, Callable[[bool], IOFormat]] = {
    '.txt' : lambda strict: IOFormat('\t', strict=strict), 
    '.tsv' : lambda strict: IOFormat('\t', strict=strict), 
    '.csv' : lambda strict: IOFormat(','), 
    '.xlsx': lambda strict: IOFormat('xlsx'),
}
_FORMAT.update({
    key[1:]: value for key, value in _FORMAT.items()
})

def get_formats(allow_excel: bool = True) -> Tuple[str]:

    """List the supported table formats.

    Parameters
    ----------
    allow_excel : bool, optional
        Whether to include XLSX formats. Default: `True`.

    Returns
    -------
    tuple
        The supported table formats.

    Examples
    --------
    >>> get_formats()
    ('.txt', '.tsv', '.csv', '.xlsx', 'txt', 'tsv', 'csv', 'xlsx')
    
    """

    if allow_excel:
        return tuple(_FORMAT)
    else:
        return tuple(key for key in _FORMAT if not key.endswith('xlsx'))
    

def format2delim(
    format: str,
    default: Optional[str] = None,
    allow_excel: bool = True,
    strict: bool = True
) -> Optional[IOFormat]:
    
    r"""Return a delimiter from its format name or extension.

    Parameters
    ----------
    format : str
        Format name.
    default : str, optional
        Default delimiter to return if delimiter not supported.
    allow_excel: bool, optional
        Whether to return 'xlsx' for Excel files. If `False`, 
        returns default or `None`. Default: `True`.
    strict : bool, optional
        Whether to allow whitespace delimiter in TSV. Default: `False`.

    Returns
    -------
    IOFormat or None
        Delimiter for TSV or CSV, or "xlsx" if Excel. If not supported
        and no default, returns None

    Examples
    --------
    >>> format2delim(".csv")
    IOFormat(in_delim=',', out_delim=',', strict=True)
    >>> format2delim("tsv")
    IOFormat(in_delim='\t', out_delim='\t', strict=True)
    >>> format2delim("tsv", strict=False)
    IOFormat(in_delim='\\s+', out_delim='\t', strict=False)
    >>> format2delim(".xlsx")
    IOFormat(in_delim='xlsx', out_delim='xlsx', strict=True)
    >>> format2delim(".cool", default=".")
    IOFormat(in_delim='.', out_delim='.', strict=True)
    >>> format2delim(".cool") is None
    True

    """
    
    if default is not None:
        default_fn = lambda strict: IOFormat(default, strict=strict)
    else:
        default_fn = lambda strict: None
    delim = _FORMAT.get(format.casefold(), default_fn)(strict)

    if delim is not None:
        if delim.in_delim == 'xlsx' and not allow_excel:
            return default_fn(strict)
    
    return delim
    

def sniff(
    file: Union[str, IO],
    default: Optional[str] = None,
    allow_excel: bool = True,
    strict: bool = True
) -> Optional[IOFormat]:

    r"""Identify the delimiter of a file from its extension.

    Parameters
    ----------
    file : str or file-like
        Input path to file or a file-like object.
    default : str, optional
        Default delimiter to return if delimiter not supported.
    allow_excel: bool, optional
        Whether to return 'xlsx' for Excel files. If `False`, 
        returns default or `None`. Default: `True`.
    strict : bool, optional
        Whether to allow whitespace delimiter in TSV. Default: `False`.

    Returns
    -------
    IOFormat or None
        Delimiter for TSV or CSV, or "xlsx" if Excel. If not supported
        and no default, returns None

    Examples
    --------
    >>> sniff("test.csv")
    IOFormat(in_delim=',', out_delim=',', strict=True)
    >>> sniff("test.tsv")
    IOFormat(in_delim='\t', out_delim='\t', strict=True)
    >>> sniff("test.tsv", strict=False)
    IOFormat(in_delim='\\s+', out_delim='\t', strict=False)
    >>> sniff("test.xlsx")
    IOFormat(in_delim='xlsx', out_delim='xlsx', strict=True)
    >>> sniff("test.cool", default=".")
    IOFormat(in_delim='.', out_delim='.', strict=True)
    >>> sniff("test.cool") is None
    True
    >>> sniff("test.xlsx")
    IOFormat(in_delim='xlsx', out_delim='xlsx', strict=True)
    >>> sniff("test.xlsx", allow_excel=False) is None
    True
    >>> sniff("test.tsv.gz")
    IOFormat(in_delim='\t', out_delim='\t', strict=True)

    """

    try:
        filename = file.name
    except AttributeError:  ## probably str
        filename = file

    if filename.endswith('.gz') or filename.endswith('.gzip'):
        new_filename, _ = os.path.splitext(filename)
        return sniff(new_filename, default, allow_excel)
    else:
        _, ext = os.path.splitext(filename)
    return format2delim(
        ext.casefold(), 
        default=default, 
        allow_excel=allow_excel,
        strict=strict,
    )


def resolve_delim(
    file: Union[str, IO],
    format: Optional[str] = None,
    default: Optional[str] = None,
    allow_excel: bool = True
) -> Optional[IOFormat]:
    
    r"""Identify the delimiter of a file.
    
    Uses the file extension, unless an explicit format is provided.

    Parameters
    ----------
    file : str or file-like
        File whose delimiter should be identified.
    format : str, optional
        Override the file extension to return a format.
    default : str, optional
        Provide this default if the extension cannot be identified,
        otherwise return `None`.
    allow_excel : bool, optional
        Whether to return 'xlsx' for Excel files. If `False`, 
        returns default or `None`. Default: `True`.

    Returns
    -------
    IOFormat or None
        Delimiter for TSV or CSV, or "xlsx" if Excel. If not supported
        and no default, returns None

    Examples
    --------
    >>> resolve_delim("test.tsv", format="csv")
    IOFormat(in_delim=',', out_delim=',', strict=True)
    >>> resolve_delim("test.tsv", format="tsv")
    IOFormat(in_delim='\t', out_delim='\t', strict=True)
    >>> resolve_delim("test.cool") is None
    True
    >>> resolve_delim("test.cool", default="\t")
    IOFormat(in_delim='\t', out_delim='\t', strict=True)

    """
    
    if format is None:
        return sniff(file, default=default, allow_excel=allow_excel)
    else:
        return format2delim(format, default=default)


def read_csv(
    filename: Union[str, TextIO], 
    rows: Optional[Union[int, Sequence[int]]] = None, 
    progress: bool = True,
    *args, **kwargs
) -> pd.DataFrame:
    
    """Read a delimited file, optionally GZIPped, optionally only specific rows.
    
    Provides a progress bar by default. Addtional arguments are passed to `pd.read_csv`.

    Parameters
    ----------
    filename : str
        Path of file to read. Optionally GZIP compressed.
    rows : list of int, optional
        Rows to read. If `None` (default), read all rows.
    progress : bool
        Whether to display a progress bar. Default: `True`.

    Returns
    -------
    pd.DataFrame
        Pandas DataFrame of the input file.

    """
    
    if rows is not None:
        rows = {0} | {row + 1 for row in rows}
        nrows = len(rows)
    else:
        nrows = 'all'

    if progress:
        print_err(f"Reading {nrows} rows from {filename}...")
    
    lines = get_lines(filename, 
                      lines=rows, 
                      progress=progress)
    return pd.read_csv(lines, *args, **kwargs)
    

def read_table(
    file: Union[str, IO],
    format: Optional[str] = None,
    progress: bool = False,
    sheet_name: Optional[Union[str, int, list]] = None,
    chunksize: Optional[int] = None,
    *args, **kwargs
) -> Union[pd.DataFrame, Iterator[pd.DataFrame]]:

    """Universal reader of tabular data files.

    Addtional arguments are passed to `read_csv` or `pd.read_excel`. If `chunksize` is
    provided, returns a iterator of chunks which is lazy for non-Excel files but greedy for 
    Excel files. Otherwise returns a DataFrame.

    Parameters
    ----------
    filename : str
        Path of file to read in CSV, TSV or Excel format. Optionally GZIP compressed.
    format : str
        Format name. Default: infer from filename
    sheet_name : str, int or list, optional
        If reading an XLSX file, which sheets to read. Default: read all sheets.

    Returns
    -------
    pd.DataFrame or iterator of pd.DataFrame
        Pandas DataFrame of the input file.
    
    """
    
    delimiter = resolve_delim(
        file, 
        format, 
        default='\t',
    )

    if delimiter.in_delim != 'xlsx':
        return read_csv(
            file, 
            sep=delimiter.in_delim,
            encoding='unicode_escape',
            progress=progress,
            chunksize=chunksize,
            *args, **kwargs
        )
    else:
        df = pd.read_excel(
            file.name, 
            engine='openpyxl',
            sheet_name=sheet_name,
            *args, **kwargs
        )
        if chunksize is None:
            return df
        else:
            return (
                df.iloc[i:(i + chunksize)] for i in range(0, df.shape[0], chunksize)
            )
        
    
def write_stream(
    df: pd.DataFrame, 
    output: Union[TextIO, str] = sys.stdout,
    format: Optional[str] = None,
    *args, **kwargs
) -> None:
    
    """Write a Pandas DataFrame to a file or stdout.
    
    Similar to pd.write_csv() but excludes the index by default and writes to 
    stdout by default with support for truncating output without complaining 
    about broken pipes.

    Addtional arguments are passed to `pd.write_csv`.

    Parameters
    ----------
    df : pd.DataFrame
        Input Pandas DataFrame to write out.
    output : str, optional
        Path to output filename. Default: stdout.
    format : str
        Format name. Default: infer from filename
    
    Returns
    -------
    None
    
    """
    
    delimiter = resolve_delim(
        output, 
        format, 
        allow_excel=False,
        default='\t',
    )
    try:
        df.to_csv(
            output,
            sep=delimiter.out_delim,
            index=False,
            *args, **kwargs
        )
    except BrokenPipeError:
        sys.exit(0)
    return None


"""Miscellaneous utilities."""

from typing import Any, Iterable, Mapping, Tuple, Optional

from functools import singledispatch
from itertools import chain

import sys

__all__ = ["colorblind_palette", "print_err", "pprint_dict", "upper_and_lower"]

_CBPAL = ('#EE7733', '#0077BB', '#33BBEE', 
          '#EE3377', '#CC3311', '#009988', 
          '#BBBBBB', "#000000")

def colorblind_palette(i: Optional[Any] = None) -> Tuple[str]:

    """Provide hexadecimal codes for a colorblind-friendly qualitative palette.

    Parameters
    ----------
    i : int, slice, iterable, optional
        Index(es) of colors to return. If not supplied, returns all.

    Returns
    -------
    tuple
        Sequence of colors encoded as hexadecimals.

    Raises
    ------
    NotImplementedError
        If `i` is not an integer, slice, or iterable.
    
    Examples
    --------
    >>> colorblind_palette(range(2))
    ('#EE7733', '#0077BB')
    >>> colorblind_palette(slice(3, 6))
    ('#EE3377', '#CC3311', '#009988')
    >>> colorblind_palette()
    ('#EE7733', '#0077BB', '#33BBEE', '#EE3377', '#CC3311', '#009988', '#BBBBBB', '#000000')
    
    """

    return _colorblind_palette(i)
    

@singledispatch
def _colorblind_palette(i: Any):
    
    try:
        return tuple(_CBPAL[j] for j in i)
    except:
        raise NotImplementedError(f"No method for colorblind_palette for type {type(i)}: {i}.")
    

@_colorblind_palette.register
def _(i: None) -> Tuple[str]:

    return _CBPAL


@_colorblind_palette.register
def _(i: slice) -> Tuple[str]:

    return _CBPAL[i]


@_colorblind_palette.register
def _(i: int) -> Tuple[str]:

    return _CBPAL[i]


def print_err(*args, **kwargs) -> None:

    r"""Print to stderr instead of stdout.

    Arguments are passed to standard `print`.

    Returns
    -------
    None

    Examples
    --------
    >>> print_err("Hello world!")  # doctest: +SKIP
    Hello world!
    >>> print_err("Hello world!", end='~\n')  # doctest: +SKIP
    Hello world!~
    >>> print_err("Hello", "world!", end='~\n')  # doctest: +SKIP
    Hello world!~

    """

    return print(*args, **kwargs, 
                 file=sys.stderr)


def pprint_dict(object: Mapping, 
                message: Optional[str] = None) -> None:
    
    """Pretty-print a dictionary to stderr, optionally prepended with a message.

    Can accept either a dictionary or an object with a `__dict__` attribute. Useful for 
    summarizing command line arguments.

    Parameters
    ----------
    object : dict or object
        Dictionary or object to print.
    message : str, optional
        Message to prepend.

    Raises
    ------
    TypeError
        If `object` is not a dictionary or not an object with a `__dict__` attribute.

    Examples
    --------
    >>> d = {"option A": 1, "option B": True, "option C": "Nothing"}  
    >>> pprint_dict(d)  # doctest: +SKIP
            option A: 1
            option B: True
            option C: Nothing
    >>> pprint_dict(d, message="Your options")  # doctest: +SKIP
    Your options:
            option A: 1
            option B: True
            option C: Nothing

    """

    if not isinstance(object, dict):
        try:
            object = vars(object)
        except TypeError:
            raise TypeError("Input to pprint_dict is not a dict and "
                            "does not have a __dict__ attribute.\n\t"
                            "Attributes: {}".format(', '.join(dir(object))))
    
    key_val_str = (f'{key}: {val}' for key, val in object.items())
    
    if message is not None:
        message = f'{message}:\n\t'
    else:
        message = '\t'

    return print_err(message + '\n\t'.join(key_val_str))


def upper_and_lower(x: Iterable[str], 
                    sort: bool = True) -> Tuple[str]:

    """Convert a list of strings into the unique set of uppercase and 
    lowercase versions.

    Useful for listing command-line options which are case-insensitive.

    Parameters
    ----------
    x : Iterable of strings
        Strings to convert.
    sort : bool, optional
        Whether to sort the output. Default: `True`

    Returns
    -------
    tuple
        Set of unique lowercase and uppercase strings.

    Examples
    --------
    >>> upper_and_lower(["A", "B"])
    ('A', 'B', 'a', 'b')
    >>> upper_and_lower(["Cat", "dog"])
    ('CAT', 'Cat', 'DOG', 'cat', 'dog')
    
    """

    upper, lower = (tuple(map(f, x)) for f in (str.upper, str.casefold))
    unique_strings = set(chain(x, upper, lower))

    if sort:
        return tuple(sorted(unique_strings))
    else:
        return tuple(unique_strings)


"""Utilities for matplotlib."""

from typing import Any, Callable, Iterable, Mapping, Tuple, Optional, Union
from functools import wraps
import os

try:
    import matplotlib.pyplot as plt
except ImportError:
    raise ImportError("Matplotlib not installed. Try installing with pip:"
                      "\n\t$ pip install matplotlib\n"
                      "\nor reinstall carabiner with matplotlib:\n"
                      "\n\t$ pip install carabiner[mpl]\n")
else:
    from matplotlib import axes, cycler, figure, rcParams, legend
    import numpy as np
    from pandas import DataFrame
from tqdm.auto import tqdm

from ..cast import cast
from ..utils import colorblind_palette as utils_colorblind_palette, print_err

TFigAx = Tuple[figure.Figure, axes.Axes]

colorblind_palette = utils_colorblind_palette

# Set default color cycle on import
rcParams['axes.prop_cycle'] = cycler(color=colorblind_palette())

def grid(
     nrow: int = 1, 
     ncol: int = 1, 
     panel_size: float = 3., 
     aspect_ratio: float = 1., 
     layout: str = 'constrained',
     sharex: Union[str, bool] = False,
     sharey: Union[str, bool] = False,
     hide_shared_ticks: bool = False,
     square: bool = False,
     *args, **kwargs
) -> TFigAx:
    
     """Create a figure and a set of subplots with sensible defaults.

     Additional arguments are passed to `matplotlib.pyplot.subplots()`.

     Parameters
     ----------
     nrow : int, optional
          Number of rows. Default: 1.
     ncol : int, optional
          Number of columns. Default: 1.
     panel_size : float, optional
          Size of panels. Default: 3.
     aspect_ratio : float
          Ratio of width over height. Default: 1 (square).
     layout : str
          Matplotlib `fig` layout. Default: "constrained".
     hide_shared_ticks : bool, optional
          Whether to hide the ticks when axes have shared scales
          from setting `sharex` or `sharey`. Default: `False`.
     square: bool
          Whether to force panels to be square. Default: `True`.

     Returns
     -------
     tuple
          Pair of `figure.Figure` and `axes.Axes` objects.
    
     """
     figsize=(
          ncol * panel_size * aspect_ratio,
          nrow * panel_size,
     )
     fig, axes = plt.subplots(
          nrow, 
          ncol, 
          figsize=figsize,
          layout=layout,
          sharex=sharex,
          sharey=sharey,
          *args, **kwargs
     )
     all_axes = fig.axes
     if square:
          for ax in all_axes:
               ax.set(aspect="equal")
     if sharex: 
          for ax in all_axes:
               ax.xaxis.set_tick_params(labelbottom=True)
     if sharey:
          for ax in all_axes:
               ax.yaxis.set_tick_params(labelleft=True)
     return fig, axes


def add_legend(
     ax: axes.Axes,
     **kwargs
) -> legend.Legend:

     """Add a legend to the right of a Matplotlib plotting axis.

     Uses a sensible default for putting the legend out of the way. Keyword arguments 
     override `loc` and `bbox_to_anchor`, and additional arguments are passed to 
     `matplotlib.axes.Axes.legend()`.

     Parameters
     ----------
     ax : matplotlib.axes.Axes
          Axes to add a legend to.

     Returns
     -------
     matplotlib.legend.Legend

     """
     default_opts = {
          "loc": 'center left',
          "bbox_to_anchor": (1, .5)
     }
     default_opts.update(kwargs)
     return ax.legend(**default_opts)


def scattergrid(
     df: DataFrame, 
     grid_columns: Union[str, Iterable[str]], 
     grid_rows: Optional[Union[str, Iterable[str]]] = None, 
     grouping: Optional[Union[str, Iterable[str]]] = None,
     log: Optional[Union[str, Iterable[str]]] = None,
     n_bins: int = 40,
     scatter_opts: Optional[Mapping[str, Any]] = None,
     hist_opts: Optional[Mapping[str, Any]] = None,
     legend_opts: Optional[Mapping[str, Any]] = None,
     *args, **kwargs
) -> TFigAx:

     """Create a scatter plot to compare sets of variables in a Pandas DataFrame.

     Similar to `pandas.plotting.scatter_matrix`, but with larger panels and control
     over which variables are log-scaled. 
     
     Additional arguments are passed to `grid()`.

     Parameters
     ----------
     df : pandas.DataFrame
          Data to plot.
     grid_columns : str | Iterable[str]
          Data columns to plot along the rows of the scatter grid. 
          Becomes the x-axes.
     grid_rows : str | Iterable[str], optional
          Data columns to plot down the columns of the scatter grid. 
          Becomes the y-axes. If not provided, uses `grid_columns` for
          all pair-wise comparison.
     grouping : str | Iterable[str], optional
          If provided, use these columns of `df` to make groups and plot each 
          data group as a differnt color.
     log : str | Iterable[str], optional
          If provided, plot these columns of `df` on a log scale.
     n_bins : int, optional
          Number of bin for histograms plotted on identity diagonal of the 
          scatter grid. Default: 40.
     scatter_opts : dict, optional
          Extra keyword arguments to pass to the Matplotlib scatter plots.
     hist_opts : dict, optional
          Extra keyword arguments to pass to the Matplotlib histogram plots.
     legend_opts : dict, optional
          Extra keyword arguments to pass to the Matplotlib legend.

     Returns
     -------
     tuple
          Pair of `figure.Figure` and `axes.Axes` objects.

     Raises
     ------
     KeyError
          If no named columns are in `df`.
    
     """

     grid_columns = [name for name in cast(grid_columns, to=list) if name in df]
     grid_rows = grid_rows or grid_columns
     grid_rows = [name for name in cast(grid_rows, to=list) if name in df]
     all_names = sorted(set(grid_columns + grid_rows))

     log = log or []
     log = [name for name in cast(log, to=list) if name in all_names]

     if grouping is None:
          grouping = "__group__"
          df = df.assign(__group__=grouping).groupby(grouping)
          dummy_group = True
     else:
          grouping = sorted(set([
               g for g in cast(grouping, to=list) if g in df
          ]))
          if len(grouping) == 0:
               raise KeyError(f"No columns in grouping ({', '.join(grouping)}) were in the DataFrame ({', '.join(df)})!")
          df = df.groupby(grouping)
          dummy_group = False

     _scatter_opts = {"s": 3.}
     _scatter_opts.update(scatter_opts or {})
     _hist_opts = {"alpha": .7} if not dummy_group else {}
     _hist_opts.update(hist_opts or {})
     _legend_opts = {}
     _legend_opts.update(legend_opts or {})

     fig, axes = grid(
          nrow=len(grid_rows), 
          ncol=len(grid_columns),
          *args, **kwargs
     )
     for axrow, grid_row_name in zip(tqdm(axes), grid_rows):
          for ax, grid_col_name in zip(axrow, grid_columns):
               make_histogram = grid_row_name == grid_col_name
               xscale = "log" if grid_col_name in log else "linear"
               yscale = "log" if (grid_row_name in log and not make_histogram) else "linear"
               ylabel = grid_row_name if not make_histogram else "Frequency"
               for group_name, group_df in df:
                    labels = {"label": ":".join(map(str, group_name))} if not dummy_group else {}
                    if make_histogram:
                         if xscale == "log":
                              values = group_df[grid_col_name].values
                              bins = np.geomspace(
                                   values.min(), 
                                   values.max(), 
                                   num=n_bins,
                              )
                         else:
                              bins = n_bins
                         ax.hist(
                              grid_col_name, 
                              data=group_df, 
                              bins=bins,
                              **_hist_opts,
                              **labels,
                         )
                    else:
                         ax.scatter(
                              grid_col_name,
                              grid_row_name, 
                              data=group_df, 
                              **_scatter_opts,
                              **labels,
                         )
               ax.set(
                    xlabel=grid_col_name, 
                    ylabel=ylabel,
                    xscale=xscale,
                    yscale=yscale,
               )
          if not dummy_group:
               add_legend(ax, **_legend_opts)
     return fig, axes


def figsaver(
     output_dir: str = ".",
     prefix: Optional[str] = None,
     dpi: int = 300, 
     format: str = 'png', 
) -> Callable[[figure.Figure, str, int, str, Optional[DataFrame]], None]:

     """Create a function to save figures in a predefined location.

     Parameters
     ----------
     output_dir : str, optional
          Directory to save figures. Default: ".".
     prefix : str, optional
          Prefix for filenames. Default: no prefix.
     dpi : int, optional
          Resolution of saved figures. Default: 300.
     format : str, optional
          File format of figures. Default: "png".

     Returns
     -------
     Callable
          A function taking Figure, name, and optionally a Pandas 
          DataFrame as arguments. Saves as {dir}/{prefix}{name}.{format}.
          If a DataFrame is provided, it as saved as {dir}/{prefix}{name}.csv.

     """

     prefix = prefix or ""
     if not os.path.exists(output_dir):
          os.mkdir(output_dir)

     def _figsave(
          fig: figure.Figure, 
          name: str, 
          df: Optional[DataFrame] = None
     ) -> None:
          """

          """
          figname = os.path.join(output_dir, f"{prefix}{name}.{format}")
          print_err(f"Saving plot at {figname}")
          fig.savefig(
               figname, 
               dpi=dpi, 
               bbox_inches='tight',
          )
          if df is not None and isinstance(df, DataFrame):
               dataname = os.path.join(output_dir, f"{prefix}{name}.csv")
               print_err(f"Saving data at {dataname}")
               df.to_csv(dataname, index=False)
          return None

     return _figsave


try:
    import tensorflow as tf
except ImportError:
    raise ImportError("\nTensorflow not installed. Try installing with pip:"
                      "\n$ pip install tensorflow\n"
                      "\nor reinstall carabiner with tensorflow:\n"
                      "\n$ pip install carabiner[deep]\n")
else:
    from tensorflow import Tensor

@tf.function(experimental_compile=True)
def sparse_matmul(a: Tensor, 
                  b: Tensor) -> Tensor:

    """Matrix multiply an indicator matrix tensor with a dense tensor.

    The indicator tensor is a `[batch x n x 1]` tensor of indices indicating 
    the single value in a row that is set to 1.

    Parameters
    ----------
    a : Tensor
        Indicator matrix tensor.
    b : Tensor
        Dense tensor.

    Returns
    -------
    Tensor

    """
    
    new_b = tf.concat([b, tf.zeros((1, b.shape[-1]))], 
                        axis=0)
    a_new = tf.where(a < 0, new_b.shape[-1] + a, a)
    a_new = tf.cast(a_new, dtype=tf.int32)

    result = tf.gather(params=new_b, 
                       indices=a_new,
                       axis=-2)
    
    return tf.reduce_sum(result, axis=-2)


@tf.function(experimental_compile=True)
def sparse_matmul_t(a: Tensor, 
                    b: Tensor) -> Tensor:

    """Matrix multiply an indicator matrix tensor with the transpose of a dense tensor.

    The indicator tensor is a `[batch x n x 1]` tensor of indices indicating 
    the single value in a row that is set to 1.

    This should be more efficient than explicitly transposing the dense tensor.

    Parameters
    ----------
    a : Tensor
        Indicator matrix tensor.
    b : Tensor
        Dense tensor to transpose.

    Returns
    -------
    Tensor

    """
    
    new_b = tf.concat([b, tf.zeros((b.shape[-2], 1))], 
                        axis=1)
    a_new = tf.where(a < 0, new_b.shape[-2] + a, a)
    a_new = tf.cast(a_new, dtype=tf.int32)

    result = tf.gather(params=new_b, indices=a_new, axis=-1)
    
    return tf.reduce_sum(result, axis=-1)


@tf.function(experimental_compile=True)
def get_param(dim0: Tensor, 
              dim1: Tensor, 
              m: Tensor) -> Tensor:

    """Matrix multiply two dense matrix tensors with a 2d dense tensor.

    The 2d dense tensor `m` is a `[batch x m x p]` tensor. The first matrix `dim0` 
    multiplies along m, and the second matrix `dim1` multiplies along p.

    Parameters
    ----------
    dim0 : Tensor
        Indicator matrix tensor.
    dim1 : Tensor
        Indicator matrix tensor.
    m : Tensor
        Dense tensor.

    Returns
    -------
    Tensor

    """

    param = dim0 @ m
    param = tf.expand_dims(dim1, axis=-2) @ tf.expand_dims(param, axis=-1)
    param = tf.squeeze(param, axis=-1)

    return param


@tf.function(experimental_compile=True)
def get_param_sparse(dim0: Tensor, 
                     dim1: Tensor, 
                     m: Tensor) -> Tensor:

    """Matrix multiply two indicator matrix tensors with a 2d dense tensor.

    The indicator tensor is a `[batch x n x 1]` tensor of indices indicating 
    the single value in a row that is set to 1.

    The 2d dense tensor `m` is a `[batch x m x p]` tensor. The first indicator `dim0` indexes
    into m, and the second indicator matrix `dim1` indexes into p.

    Parameters
    ----------
    dim0 : Tensor
        Indicator matrix tensor.
    dim1 : Tensor
        Indicator matrix tensor.
    m : Tensor
        Dense tensor.

    Returns
    -------
    Tensor

    """

    dim0, dim1 = tf.cast(dim0, dtype=tf.int32), tf.cast(dim1, dtype=tf.int32)

    new_m = tf.concat([m, tf.zeros((m.shape[-2], 1))], 
                      axis=1)
    new_m = tf.concat([new_m, 
                       tf.zeros((1, new_m.shape[-1]))], 
                       axis=0)
    dim0 = tf.where(dim0 < 0, new_m.shape[-2] + dim0, dim0)
    dim1 = tf.where(dim1 < 0, new_m.shape[-1] + dim1, dim1)

    param = tf.gather_nd(new_m, 
                         indices=tf.concat([dim0, dim1], 
                         axis=-1))[:, tf.newaxis]

    return param


"""Utilities for IO."""

from typing import Callable, Iterator, Optional, Sequence, Tuple, Union
from functools import partial
import gzip
from io import StringIO, TextIOWrapper
import math
from tempfile import _TemporaryFileWrapper

from ..itertools import tenumerate
from ..utils import print_err

def _enumerate_file(filename: str, 
                    opener: Optional[Callable[[str], TextIOWrapper]] = None, 
                    progress: bool = True,
                    total: Optional[int] = None,
                    *args, **kwargs) -> Iterator[Tuple[int, str]]:
    
    if isinstance(filename, str):
        if opener is None:
            if filename.endswith('.gz'):
                opener = partial(gzip.open, mode='rt')
            else:
                opener = open
        
        handle = opener(filename, *args, **kwargs)
    elif isinstance(filename, TextIOWrapper) or isinstance(filename, _TemporaryFileWrapper):
        handle = filename
    else:
        raise IOError(f"Object {filename} is not a string (path) or a file-like object.")
    
    enumerator = (partial(tenumerate, total=total) if progress 
                  else enumerate)

    return enumerator(handle)
    

def count_lines(filename: Union[TextIOWrapper, str], 
                progress: bool = True,
                *args, **kwargs) -> int:
    
    """Count lines in a file, optionally gzipped.

    Provides a progress bar by default.
    
    Parameters
    ----------
    filename : str
        Path of file to read. Optionally GZIP compressed.
    progress : bool
        Whether to display a progress bar. Default `True`.

    Returns
    -------
    int
        Number of lines in input file.
    
    """
    
    if progress:
        print_err(f"Counting lines in {filename}...")

    for i, _ in _enumerate_file(filename, 
                                progress=progress, 
                                *args, **kwargs):

            pass
                
    return i + 1


def get_lines(filename: Union[TextIOWrapper, str],
              lines: Optional[Union[int, Sequence[int]]] = None,
              progress: bool = True,
              outfile: Optional[TextIOWrapper] = None,
              *args, **kwargs) -> Union[TextIOWrapper, StringIO]:
    
    """Extract lines from a file, optionally GZIPped, by line number.

    Provides a progress bar by default.

    Parameters
    ----------
    filename : str or file-like
        Path of file to read, or a file-like object. Optionally GZIP compressed.
    lines : list of int, optional
        Rows to read. If `None` (default), read all rows.
    progress : bool
        Whether to display a progress bar. Default `True`.
    outfile : file-like, optional
        Open file handle for output.

    Returns
    -------
    TextIOWrapper or StringIO
        File-like object containing lines from the input file.
    
    """

    outfile = outfile or StringIO()
    
    if lines is not None:
        
        if isinstance(lines, int):
            lines = range(lines)

        line_numbers_to_keep = set(lines)
        nlines_to_read = max(line_numbers_to_keep)

    else:

        line_numbers_to_keep = set()
        nlines_to_read = math.inf

    for i, line in _enumerate_file(filename, 
                                   progress=progress, 
                                   total=nlines_to_read,
                                   *args, **kwargs):

        if (lines is None) or (i in line_numbers_to_keep):

            print(line, file=outfile, end='')

        if i > nlines_to_read:

            break

    outfile.seek(0)  ## Essential to return to start of file 

    return outfile


"""Decorators to faciltiate functional programming."""

from typing import Any, Callable, Iterable, Optional

from functools import wraps

from .cast import cast, flatten

TDecorator = Callable[[Callable, Any], Callable]

def decorator_with_params(decorator: TDecorator) -> TDecorator:

    """Convert a decorator to be used with optional parameters.

    Can itself be used as a decorator.

    The input decorator should take a function as its first argment, and
    further optional parameters. This `decorator_with_params` will 
    return a decorator that can be used with or without specifying the optional
    parameters. (Normally this a very convoluted code.)

    Parameters
    ----------
    decorator : Callable
        Decorator to convert.

    Returns
    -------
    Callable
        Converted decorator.

    Examples
    --------
    >>> def decor(f, suffix="World"): return lambda x: f(x + suffix)
    ...
    >>> @decor
    ... def printer(x): print(x)
    ... 
    >>> @decor(suffix="everyone") 
    ... def printer2(x): print(x)
    ... 
    Traceback (most recent call last):
        ...
    TypeError: decor() missing 1 required positional argument: 'f'
    >>> decor2 = decorator_with_params(decor)
    >>> @decor2(suffix="everyone")
    ... def printer3(x): print(x)
    ... 
    >>> printer("Hello ")
    Hello World
    >>> printer3("Hello ")
    Hello everyone

    """

    @wraps(decorator)
    def _parameterized_decorator(decorated_function: Optional[Callable] = None,
                                 *args, **kwargs) -> Callable:

        @wraps(decorated_function)
        def _decorated_function(f):

            return decorator(f, *args, **kwargs)

        return (decorator(decorated_function,  *args, **kwargs) 
                if callable(decorated_function) 
                else _decorated_function)

    return _parameterized_decorator


@decorator_with_params
def return_none_on_error(f: Callable, 
                         exception: Optional[Exception] = None) -> Callable:
    
    """Force a function to return None instead of raising an exception.

    Parameters
    ----------
    f : Callable, optional
        Function to convert. If not provided, returns a decorator.
    exception : Exception, optional
        Exception type to bypass. Default: all exceptions.

    Returns
    -------
    Callable
        If f is provided, then decorated f. Otherwise returns a decorator.

    Examples
    --------
    >>> def error_maker(x): raise KeyError
    ... 
    >>> @return_none_on_error
    ... def error_maker2(x): raise KeyError
    ... 
    >>> @return_none_on_error(exception=ValueError)
    ... def error_maker3(x): raise KeyError
    ... 
    >>> error_maker('a')  
    Traceback (most recent call last):
        ...
    KeyError
    >>> error_maker2('a')
    >>> error_maker3('a') 
    Traceback (most recent call last):
        ...
    KeyError

    """
    
    exception = exception or Exception
    
    @wraps(f)
    def wrapped_function(*args, **kwargs):

        try:
            return f(*args, **kwargs)
        except exception:
            return None

    return wrapped_function


def vectorize(f: Callable) -> Callable:

    """Convert a scalar function into a vectorized mapping function.

    The converted function returns a lazy iterable if passed an iterable, 
    and a scalar if passed a scalar.

    Parameters
    ----------
    f : Callable
        Scalar function to convert.

    Returns
    -------
    Callable
        Vectorized function.

    Examples
    --------
    >>> @vectorize
    ... def vector_adder(x): return x + 1
    ...
    >>> list(vector_adder(range(3)))
    [1, 2, 3]
    >>> list(vector_adder((4, 5, 6)))
    [5, 6, 7]
    >>> vector_adder([10])
    11
    >>> vector_adder(10)
    11

    """

    @wraps(f)
    def vectorized_function(arg0: Any, 
                            *args, **kwargs) -> Iterable:

       arg0 = cast(arg0, to=list)

       return flatten(f(_x, *args, **kwargs) for _x in arg0)
        
    return vectorized_function




"""Utilities for iterables."""

from typing import Any, Iterable

from itertools import islice

from tqdm.auto import tqdm

def batched(iterable: Iterable[Any], 
            n: int) -> Iterable[Iterable[Any]]:

    """Copy of the Python function for versions where it's not available.

    # batched('ABCDEFG', 3) --> ABC DEF G

    Parameters
    ----------
    iterable : Iterable
        Iterable to batch.
    n : int
        Batch size.

    Returns
    -------
    Iterable
        Iterable of iterables with size `n`.

    Examples
    --------
    >>> for batch in batched('abcdef', 3): print(list(batch))
    ... 
    ['a', 'b', 'c']
    ['d', 'e', 'f']
    
    """
   
    if n < 1:
        raise ValueError('n must be at least one')
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch


def tenumerate(x: Iterable[Any], 
               *args, **kwargs) -> Iterable[Any]:
    
    """Standard `enumerate` but with a progress bar.

    Additional arguments are passed to `tqdm.tqdm`.

    Parameters
    ----------
    iterable : Iterable
        Iterable to enumerate.

    Returns
    -------
    Iterable
        Iterable of tuples containing index and item.
    
    """

    return enumerate(tqdm(x, *args, **kwargs))


"""Tools for randomness."""

from typing import Iterable, List

from math import log, log1p, floor
from random import random, randrange, shuffle
from itertools import islice


def sample_iter(iterable: Iterable, 
                k: int = 1, 
                shuffle_output:bool = True) -> List:

    """Random sampling from an iterable of unknown length without needing it to be held entirely in
    memory.

    This allows sampling from an iterable without needing the iterable to be held entirely in
    memory at one time and without knowing the length of the iterable ahead of
    time. A typical case might be selecting lines from a large text file while
    reading the file in a single pass.

    One limitation is that the sampled population must fit in memory.

    Based in Reservoir Sampling: https://en.wikipedia.org/wiki/Reservoir_sampling

    Originally written here: https://bugs.python.org/issue41311

    Based on GitHub Gist: https://gist.github.com/oscarbenjamin/4c1b977181f34414a425f68589e895d1
    
    Parameters
    ----------
    iterable : Iterable
        Sequence of items which may or may not be held in memory.
    k : int, optional
        Number of items to sample. Default: 1.
    shuffle_output : bool, optional
        Return items in random order (default). Otherwise return in
        original order.

    Returns
    -------
    list
        Sequence sampled from `iterable`.

    Examples
    --------
    >>> from string import ascii_letters
    >>> from itertools import chain
    >>> from random import seed
    >>> seed(1)
    >>> sample_iter(chain.from_iterable(ascii_letters for _ in range(1000000)), 10)
    ['X', 'c', 'w', 'q', 'T', 'e', 'u', 'w', 'E', 'h']
    >>> seed(1)
    >>> sample_iter(chain.from_iterable(ascii_letters for _ in range(1000000)), 10, shuffle_output=False)
    ['T', 'h', 'u', 'X', 'E', 'e', 'w', 'q', 'c', 'w']

    """

    iterator = iter(iterable)
    values = list(islice(iterator, k))

    irange = range(len(values))
    indices = dict(zip(irange, irange))

    kinv, W = 1. / k, 1.

    while True:

        W *= random() ** kinv

        # random() < 1.0 but random() ** kinv might not be
        # W == 1.0 implies "infinite" skips
        if W >= 1.:
            break

        # skip is geometrically distributed with parameter W
        skip = floor( log(random())/log1p(-W) )
        try:
            newval = next(islice(iterator, skip, skip+1))
        except StopIteration:
            break

        # Append new, replace old with dummy, and keep track of order
        remove_index = randrange(k)
        values[indices[remove_index]] = None
        indices[remove_index] = len(values)
        values.append(newval)

    values = [values[indices[i]] for i in irange]

    if shuffle_output:
        shuffle(values)

    return values


"""Command-line interface for carabiner."""

# from typing import Sequence
from argparse import FileType, Namespace
import sys

from .cliutils import clicommand, CLIOption, CLICommand, CLIApp

__version__ = '0.0.3.post1'

@clicommand(message='Parsed arguments', name='cbnr')
def _main(args: Namespace) -> None:
    
    return None


def main() -> None:

    inputs = CLIOption('inputs',
                        type=str,
                        default=[],
                        nargs='*',
                        help='')
    output = CLIOption('--output', '-o', 
                        type=FileType('w'),
                        default=sys.stdout,
                        help='Output file. Default: STDOUT')
    formatting = CLIOption('--format', '-f', 
                           type=str,
                           default='TSV',
                           choices=['TSV', 'CSV', 'tsv', 'csv'],
                           help='Format of files. Default: %(default)s')

    test = CLICommand("test",
                      description="Test CLI subcommand using Carabiner utilities.",
                      options=[inputs, output, formatting],
                      main=_main)

    app = CLIApp("Carabiner", 
                 version=__version__,
                 description="Test CLI app using Carabiner utilities.",
                 commands=[test])

    app.run()

    return None


if __name__ == '__main__':

    main()




from .utils import (
    colorblind_palette, 
    print_err, 
    pprint_dict, 
    upper_and_lower,
)
from .cast import cast
from .cliutils import (
    clicommand, 
    CLIOption, 
    CLICommand, 
    CLIApp,
)
from .decorators import (
    decorator_with_params,
    return_none_on_error,
    vectorize,
)
from . import (
    collections, 
    itertools, 
    random,
)


"""Tweaked casting between Python types."""

from typing import Any, List, TextIO, Union

from collections.abc import Iterable

from functools import singledispatch
from itertools import chain
import gzip
from io import StringIO, TextIOWrapper

from .itertools import batched


@singledispatch
def clist(x) -> List:

    """Cast an object to a list.

    Parameters
    ----------
    x : str or iterable
        Object to cast.

    Returns
    -------
    list

    Examples
    --------
    >>> clist('Hello world')
    ['Hello world']
    >>> clist(('Hello', 'world'))  # doctest: +SKIP
    ['Hello', 'world']
    
    """

    try:
        return list(x)
    except TypeError:  ## if x is not iterable
        return [x]


@clist.register
def _(x: str) -> List[str]:

    return [x]


@clist.register
def _(x: int) -> List[int]:

    return [x]


@clist.register
def _(x: float) -> List[float]:

    return [x]
    
    
@clist.register
def _(x: TextIOWrapper) -> List[TextIOWrapper]:

    return [x]
    

@singledispatch
def cio(x, *args, **kwargs) -> TextIOWrapper:

    r"""Cast an object to a file-like object.

    Addtional parameters are passed to `open` or `gzip.open`.

    If the input is a string, then it is treated as a filename. If it ends
    in `.gz` or `.gzip`, then it is assumed to be a GZIP-compressed file.

    If the input is an open file, then the file is closed and reopened with 
    the provided parameters. This can be useful to close for writing and 
    open for reading.

    If the input is a `io.StringIO` object, then the cursor is moved to the
    start and the object is returned.

    Parameters
    ----------
    x : str or iterable
        Object to cast.

    Returns
    -------
    TextIOWrapper

    Examples
    --------
    >>> cio('Hello.txt')  # doctest: +SKIP
    <_io.TextIOWrapper name='Hello.txt' mode='r' encoding='UTF-8'>
    >>> cio('Hello.txt', 'w')
    <_io.TextIOWrapper name='Hello.txt' mode='w' encoding='UTF-8'>
    >>> list(cio(StringIO("Hello\nworld")))
    ['Hello\n', 'world']

    """

    raise TypeError(f"Can't force object of type {type(x).__name__} "
                     "to be a file-like")


@cio.register
def _(x: str, *args, **kwargs):

    if x.endswith('.gz') or x.endswith('.gzip'):

        return gzip.open(x, *args, **kwargs)

    else:

        return open(x, *args, **kwargs)


@cio.register
def _(x: TextIOWrapper, *args, **kwargs):

    if x.name in ('<stdin>', '<stdout>'):

        return x

    else:

        return cio(x.name, *args, **kwargs)


@cio.register
def _(x: StringIO, *args, **kwargs):

    x.seek(0)

    return x


@singledispatch
def cstr(x) -> str:

    """Cast an object to a string.

    If the object is a file-like object, return the filename.

    Parameters
    ----------
    x 
        Object to cast.

    Returns
    -------
    str

    Examples
    --------
    >>> cstr(["Hello", "world"])
    "['Hello', 'world']"
    >>> cstr(open("test/test-file.txt", "w"))
    'test/test-file.txt'
    
    """

    return str(x)


@cstr.register
def _(x: TextIOWrapper):

    return x.name   


CASTING_FUNCTIONS = dict(str=cstr, 
                         list=clist, 
                         TextIO=cio,
                         TextIOWrapper=cio)

def _cast_base(object: Any,
               to: str,
               *args, **kwargs):
    try:
        casting_function = CASTING_FUNCTIONS[to]
    except KeyError:
        raise TypeError(f"Cannot cast {type(object).__name__} to {to}.")
    return casting_function(object, 
                            *args, **kwargs)


@singledispatch
def _cast(to, 
          object: Any, 
          *args, **kwargs) -> Any:
    
    return _cast_base(object, str(to),
                      *args, **kwargs)


@_cast.register
def _(to: type, 
      object: Any, 
      *args, **kwargs) -> Any:
    
    return _cast_base(object, to.__name__,
                      *args, **kwargs)


@_cast.register
def _(to: str, 
      object: Any, 
      *args, **kwargs) -> Any:
    
    return _cast_base(object, to,
                      *args, **kwargs)
        

def cast(object: Any, 
         to: Union[str, type], 
         *args, **kwargs) -> Any:

    """Cast an object to another type.

    Parameters
    ----------
    x 
        Object to cast.
    to : str, type
        Type to cast to. Available: `str`, `list`, `TextIO`, `TextIOWrapper`

    Returns
    -------
        Casted object.

    Raises
    ------
    TypeError
        If the object cannot be casted to the type.

    Examples
    --------
    >>> cast("abc", list)
    ['abc']
    >>> cast(["abc"], 'str')
    "['abc']"

    """

    return _cast(to, object=object, 
                 *args, **kwargs)
    

@singledispatch
def flatten(x) -> Any:

    """Return the only item for iterables of length one.

    If `x` is a lazy iterator, then this returns a lazy iterator if more
    than one item in `x`.

    Parameters
    ----------
    x : str, iterable
        Object to flatten.

    Returns
    -------
    iterable or item
    
    Raises
    ------
    TypeError
        If `x` is not a supported type.

    Examples
    --------
    >>> flatten("Hello")
    'Hello'
    >>> flatten(["Hello"])
    'Hello'
    >>> flatten(["Hello", "world"])
    ['Hello', 'world']
    >>> flatten(tuple('abc'))
    ('a', 'b', 'c')
    >>> list(flatten(range(3)))
    [0, 1, 2]
    >>> flatten(range(1))
    0
    
    """

    raise TypeError(f"Cannot flatten object of type {type(x).__name__}.")


@flatten.register
def _(x: list) -> Any:

    if len(x) == 1:

        return x[0]
    
    else:

        return x
    

@flatten.register
def _(x: str) -> Any:

    return x


@flatten.register
def _(x: int) -> Any:

    return x


@flatten.register
def _(x: float) -> Any:

    return x


    
@flatten.register
def _(x: tuple) -> Any:

   return tuple(flatten(list(x)))
    

@flatten.register
def _(x: set) -> Any:

    if len(x) == 1:

        return list(x)[0]
    
    else:

        return x


@flatten.register
def _(x: Iterable) -> Any:

    iterators = []
    make_list = True

    for i, batch in enumerate(batched(x, 1000)):
        
        if make_list:
            batch = list(batch)

        if i == 0:
            
            if len(batch) == 1:

                return batch[0]

            else:

                make_list = False

        iterators.append(batch)

    return chain.from_iterable(iterators) 



"""Utilities for constructing command-line interfaces."""

from typing import Callable, Iterable, Mapping, Optional

from argparse import ArgumentParser, Namespace, RawDescriptionHelpFormatter
from dataclasses import asdict, dataclass, field
from datetime import timedelta
from functools import wraps
import textwrap
from time import time

from .decorators import decorator_with_params
from .utils import print_err, pprint_dict

TMainFunction = Callable[[Namespace], None]

@decorator_with_params
def clicommand(main: TMainFunction,
               message: Optional[str] = None, 
               name: Optional[str] = None) -> TMainFunction:
    
    """Convert a function to act informatively as the main function in a CLI app.

    Parameters
    ----------
    message : str, optional
        Message to prepend to reporting of parameters.
    name: str, optional
        Name of the app.

    Returns
    -------
    Callable
        Decorator for function to be used as a main function in a CLI app.

    Examples
    --------
    >>> from argparse import Namespace
    >>> def main_func(args): print("Hello world!")
    >>> args = Namespace(a=1, b="Testing")
    >>> clicommand(main_func, name="Main program")(args)  # doctest: +SKIP
    🚀 Processing with the following parameters:
            a: 1
            b: Testing
    Hello world!
    ⏰ Completed Main program in 0:00:00.000132
    
    """
    
    message = message or "Processing with the following parameters"
    name = name or "process"
    
    @wraps(main)
    def _clicommand(args: Namespace) -> None:


        start_time = time()

        pprint_dict(args,
                    message="🚀 " + message)

        main(args)

        process_time = time() - start_time

        print_err(f'⏰ Completed {name} in {timedelta(seconds=process_time)}')

        return None

    return _clicommand
    

@dataclass
class BaseCLIOption:

    """Store command-line option parameters.

    Takes the same parameters as `argparse.ArgumentParser().add_arguments()`, but 
    automatically appends defaults or "Required" to help string if not already included.

    """

    args: Optional[Iterable] = field(default_factory=list)
    kwargs: Optional[Mapping] = field(default_factory=dict)

    def __post_init__(self):

        if 'required' in self.kwargs and 'default' in self.kwargs:

            if self.kwargs['default'] is not None and self.kwargs['required']:

                raise AttributeError(f"Cannot set both {self.kwargs['required']=} and "
                                     f"{self.kwargs['default']=} for CLIOption.")

        if 'help' in self.kwargs:

            if ('default' in self.kwargs and 
                ' Default: ' not in self.kwargs['help'] and 
                self.kwargs['default'] is not None):

                self.kwargs['help'] += ' Default: "{}".'.format(self.kwargs['default'])

            elif 'required' in self.kwargs and self.kwargs['required']:

                self.kwargs['help'] += ' Required.'


class CLIOption(BaseCLIOption):

    """Store command-line option parameters.

    Takes the same parameters as `argparse.ArgumentParser().add_arguments()`, but 
    automatically appends defaults or "Required" to help string if not already included.

    Provides a `replace` method so that options can be re-used but with a few
    parameters altered.

    """

    def __init__(self, *args, **kwargs):

        super().__init__(args=args, 
                         kwargs=kwargs)

    def replace(self, **kwargs):

        """Substitute named parameters and return a new `CLIOPtion`
        object

        """

        _copy = BaseCLIOption(**asdict(self))

        for key, value in kwargs.items():

            _copy.kwargs[key] = value

        return _copy


@dataclass
class CLICommand:

    """Store parameters for a command-line app's subcommands.

    Parameters
    ----------
    name : str
        Subcommand name.
    description : str, optional
        Help string.
    options : Iterable[CLIOption]
        Iterable of option objects.
    main: Callable
        Function to run when this command is called.

    """

    name: str
    description: Optional[str] = None
    options: Iterable[CLIOption] = field(default_factory=list)
    main: Optional[Callable] = None


@dataclass
class CLIApp:

    """A command-line app.

    Container for subcommands, which in turn contain options. Once assembled,
    the app can be run with the `run()` method.

    Parameters
    ----------
    name : str
        Name of the app.
    version : str
        Version of the app.
    description : str
        Help string.
    commands : Iterable[CLICommand]
        Set of subcommands.

    """

    name: str
    version: str
    description: Optional[str] = None
    commands: Iterable[CLICommand] = field(default_factory=list)
    _parser: ArgumentParser = field(init=False)
    _subcommands: Iterable = field(init=False, default_factory=list)
    
    def __post_init__(self):

        self._parser = ArgumentParser(formatter_class=RawDescriptionHelpFormatter,
                                      description=textwrap.dedent(self.description))
        self._parser.add_argument("--version", "-v",
                                  action="version",
                                  version='%(prog)s {version}'.format(version=self.version))

        if len(self.commands) > 0:

            subcommands =  self._parser.add_subparsers(title='Sub-commands', 
                                                       dest='subcommand',
                                                       help='Use these commands to specify the tool you want to use.')
        
            for command in self.commands:

                this_command = subcommands.add_parser(command.name, 
                                                      help=command.description)
                this_command.set_defaults(func=command.main)

                for option in command.options:

                    this_command.add_argument(*option.args, **option.kwargs)

                self._subcommands.append(this_command)

        self._args = self._parser.parse_args()

    def run(self) -> None:

        """Run the app.

        """

        return self._args.func(self._args)




"""Useful data structures."""

from collections import UserDict
from itertools import chain

class MultiKeyDict(UserDict):

    """Dictionary where multiple keys can be accessed at once.

    Examples
    --------
    >>> d = MultiKeyDict(a=1, b=2, c=3)
    >>> d
    {'a': 1, 'b': 2, 'c': 3}
    >>> d['c']
    {'c': 3}
    >>> d['a', 'b']
    {'a': 1, 'b': 2}

    """

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)


    def __getitem__(self, *args, **kwargs):

        args = chain.from_iterable(args)

        return {arg: self.data[arg] for arg in args}        